/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * dyn_lwp.C -- cross-platform segments of the LWP handler class
 * $Id: dyn_lwp.C,v 1.4 2002/12/20 07:49:56 jaw Exp $
 */

#include "common/h/headers.h"
#include "dyn_lwp.h"
#include "process.h"
#include <assert.h>

dyn_lwp::dyn_lwp() :
  proc_(NULL),
  lwp_(0),
  fd_(0),
  hw_previous_(0),
  sw_previous_(0),
  stoppedInSyscall_(false),
  postsyscallpc_(0)
{
};

dyn_lwp::dyn_lwp(unsigned lwp, process *proc) :
  changedPCvalue(0),
  proc_(proc),
  lwp_(lwp),
  fd_(0),
  hw_previous_(0),
  sw_previous_(0),
  stoppedInSyscall_(false),
  postsyscallpc_(0)

{
}

dyn_lwp::dyn_lwp(unsigned lwp, handleT fd, process *proc) :
  changedPCvalue(0),
  proc_(proc),
  lwp_(lwp),
  fd_(fd),
  hw_previous_(0),
  sw_previous_(0),
  stoppedInSyscall_(false),
  postsyscallpc_(0)

{
}

dyn_lwp::dyn_lwp(const dyn_lwp &l) :
  changedPCvalue(0),
  proc_(l.proc_),
  lwp_(l.lwp_),
  fd_(0),
  hw_previous_(0),
  sw_previous_(0),
  stoppedInSyscall_(false),
  postsyscallpc_(0)

{
}

dyn_lwp::~dyn_lwp()
{
  if (fd_) closeFD();
}

// Not sure this is a good idea... when would we be walking the stack
// (conceptually) of an LWP rather than the thread running on it?
// For now: non-MT will walk getDefaultLWP() since it doesn't understand
// multithreaded programs

// stackWalk: return parameter.
bool dyn_lwp::walkStack(pdvector<Frame> &stackWalk)
{
  // We cheat (a bit): this method is here for transparency, 
  // but the process class does the work in the walkStackFromFrame
  // method. We get the active frame and hand off.
  Frame active = getActiveFrame();
  
  return proc_->walkStackFromFrame(active, stackWalk);
}




